/*
   Copyright (C) 2003 Paul Brossier

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*/

#include "aubio_priv.h"
#include "sample.h"
#include "mathutils.h"
#include "biquad.h"
#include "peakpick.h"

/* parms: peak picking parameters, default values in brackets
 * sparms(thresh,alpha,cutoff,window,plotf,threshfn,picker)
 *
 *	   [<----post----|--pre-->]
 *	.................|.............
 *	time->           ^now
 */
struct _pickparams_t {
	/** thresh: offset threshold [0.033 or 0.01] */
	smpl_t threshold; 	
	/** win_post: median filter window length (causal part) [8] */
	uint_t 	win_post; 			
	/** pre: median filter window (anti-causal part) [post-1] */
	uint_t 	win_pre; 				
	/**	threshfn: name or handle of fn for computing adaptive threshold [@median]  */
	//aubio_thresholdfn_t thresholdfn;
	/**	picker:   name or handle of fn for picking event times [@peakpick] */
	aubio_pickerfn_t pickerfn;

	/** biquad lowpass filter */
	aubio_biquad_t * biquad;
	/** original onsets */
	fvec_t * onset_keep;
	/** modified onsets */
	fvec_t * onset_proc;
	/* peak picked window [3] */
	fvec_t * onset_peek;
	/** scratch pad for biquad and median */
	fvec_t * scratch;

	/** \bug should be used to calculate filter coefficients */
	/* cutoff: low-pass filter cutoff [0.34, 1] */
	/* smpl_t cutoff; */

	/* not used anymore */
	/* tolerance [0.05] */
	/* smpl_t tol; 			*/
	/* time precision [512/44100  winlength/samplerate, fs/buffer_size */
	/* smpl_t tau; */

	/* alpha: normalisation exponent [9] */
	/* smpl_t alpha; */
};


/* modified version of samer pp, based on mean  rather than alpha norm and min
*/
uint_t aubio_peakpick_pimrt(fvec_t * onset,  pickparams_t * p) {
	fvec_t * onset_keep = (fvec_t *)p->onset_keep;
	fvec_t * onset_proc = (fvec_t *)p->onset_proc;
	fvec_t * onset_peek = (fvec_t *)p->onset_peek;
	fvec_t * scratch    = (fvec_t *)p->scratch;
	smpl_t mean = 0., median = 0.;
	uint_t length = p->win_post + p->win_pre + 1;
	uint_t i = 0, j;

	/* store onset in onset_keep */
	/* shift all elements but last, then write last */
	/* for (i=0;i<channels;i++) { */
	for (j=0;j<length-1;j++) {
		onset_keep->data[i][j] = onset_keep->data[i][j+1];
		onset_proc->data[i][j] = onset_keep->data[i][j];
	}
	onset_keep->data[i][length-1] = onset->data[i][0];
	onset_proc->data[i][length-1] = onset->data[i][0];
	/* } */

	/* filter onset_proc */
	/** \bug filtfilt calculated post+pre times should be only once !? */
	/* if (p->cutoff<1) { */
	aubio_biquad_do_filtfilt(p->biquad,onset_proc,scratch);
	//aubio_biquad_do(p->biquad,onset_proc);
	//aubio_lowpassresoiir(onset_proc,p->cutoff,p->gain);
	/* } */

	/* calculate mean and median for onset_proc */
	/* for (i=0;i<onset_proc->channels;i++) { */
	mean = vec_mean(onset_proc);
	/* copy to scratch */
	for (j = 0; j < length; j++)
		scratch->data[i][j] = onset_proc->data[i][j];
	median = vec_median(scratch);
	/* } */

	/* for (i=0;i<onset->channels;i++) { */
	/* shift peek array */
	for (j=0;j<3-1;j++) 
		onset_peek->data[i][j] = onset_peek->data[i][j+1];
	/* calculate new peek value */
	onset_peek->data[i][2] = 
		onset_proc->data[i][p->win_post] - median - mean * p->threshold;
	/* } */
	//AUBIO_DBG("%f\n", onset_peek->data[0][2]);
	return (p->pickerfn)(onset_peek,1);
}


pickparams_t * new_aubio_peakpicker(smpl_t threshold) {
	pickparams_t * t = AUBIO_NEW(pickparams_t);
	t->threshold = 0.1;/* 0.0668; 0.33; 0.082; 0.033; */
	if (threshold > 0. && threshold < 10.)
		t->threshold = threshold;
	t->win_post = 5;
	t->win_pre = 1;

	//t->thresholdfn = (aubio_thresholdfn_t)(vec_median); /* (vec_mean); */
	t->pickerfn = (aubio_pickerfn_t)(vec_peakpick);

	t->scratch = new_fvec(t->win_post+t->win_pre+1,1);
	t->onset_keep = new_fvec(t->win_post+t->win_pre+1,1);
	t->onset_proc = new_fvec(t->win_post+t->win_pre+1,1);
	t->onset_peek = new_fvec(3,1);

	/* cutoff: low-pass filter cutoff [0.34, 1] */
	/* t->cutoff=0.34; */
	t->biquad = new_aubio_biquad(0.1600,0.3200,0.1600, -0.5949, 0.2348);

	return t;

	/* tolerance [0.05] */
	/* t->tol=0.05; *//* not used here */
	/* time precision [256/22050  winlength/samplerate, fs/buffer_size */
	/* t->tau = 256/22050; *//* not used here */				
	/* t->gain=1.; */
	/* alpha: normalisation exponent [9] */
	/** \bug above 8, goes out of bounds **/
	/* t->alpha=7; *//* not used anymore */		
}

/** Samer's Peaking picking
 *  \param onset detection function (sampling rate assumed 22050/256 Hz)
 *  \param parms peak-picking parameters (see SPARMS)
 * \bug broken -> use pimrt for real time instead
 */
smpl_t aubio_peakpick_samerpp(fvec_t * onset, pickparams_t * p) {
	/* as found by JPB 28/10/2003 */
	/* remove DC component */
	//smpl_t alphanorm = 0;
	//smpl_t min = vec_min(onset);
	smpl_t athres = 0.;
	//smpl_t tmpmax = vec_max(onset);
	uint_t i, j;

	/*
	   p->maxthreshold = (p->maxthreshold  > tmpmax ) ? p->maxthreshold : tmpmax;
	   p->maxthreshold = (p->maxthreshold  > 2.*tmpmax ) ? tmpmax : p->maxthreshold;
	   */
	/*
	   if (alphanorm == 0.) {
	   alphanorm = 0.001;
	   }
	   */

	/*
	   for (i=0;i<onset->channels;i++) {
	   for (j = 0; j < onset->length; j++) {
	//brute force replacement for alpha norm
	onset->data[i][j] /= 1500.;
	onset->data[i][j] -= min/1500.;
	}
	}
	*/
	/* alpha norm skipped 
	   alphanorm = vec_alpha_norm(onset,p->alpha);
	   for (i=0;i<onset->channels;i++) {
	   for (j = 0; j < onset->length; j++) {
	   onset->data[i][j] /= alphanorm;
	   }
	   }
	   */
	/* low pass filtering */

	/* if (p->cutoff<1) { */
	aubio_biquad_do_filtfilt(p->biquad,p->onset_proc,p->scratch);
	/* } */
	// compute adaptive threshold 
	//onset=onset-p->thresh*vec_alpha_norm(onset,p->alpha);
	//onset=onset-p->thresh*cauchy_norm(onset,0.3,1000);
	for (i=0;i<onset->channels;i++) {
		for (j = 0; j < onset->length; j++) {
			p->scratch->data[i][j] = onset->data[i][j];
		}
		//athres = (p->thresholdfn)(p->scratch);
		//calculate median on the copy
		athres = vec_median(p->scratch);
		for (j = 0; j < onset->length; j++) {
			// bug, should be median
			//fn_threshfn((p->threshfn),onset,p->post,p->pre);
			onset->data[i][j] -= athres;
			onset->data[i][j] -= p->threshold;// * p->maxthreshold;
		}
	}
	AUBIO_DBG("%f\n", onset->data[0][p->win_post-1]);
	//AUBIO_DBG("%f\n", athres);
	/* don't trust this filter_ioi function */
	//T=tau*filter_ioi(thresh_peaks(onset,thr+thresh)-1,onset,2);
	return (p->pickerfn)(onset, p->win_post-1);
}

